package Src;

import java.util.ArrayList;
import java.util.Date;

public class Algo {
	
	//methode pour transformer des coordonnées en latitude/longitude en distance en km
	    public static double degreesToKm(double lat1,double lon1,double lat2,double lon2) {
	  	  int R = 6371; // rayon de la terre
	  	  
	  	  double dLat = degToRad(lat2-lat1);
	  	  double dLon = degToRad(lon2-lon1);
	  	  lat1 = degToRad(lat1);
	  	  lat2 = degToRad(lat2);
	
	  	  double a = Math.sin(dLat/2) * Math.sin(dLat/2) +
	  	          Math.cos(lat1) * Math.cos(lat2) * 
	  	          Math.sin(dLon/2) * Math.sin(dLon/2);
	  	  double c = 2 * Math.atan2(Math.sqrt(a), Math.sqrt(1-a));
	  	  return R*c;
	  	}
	  
      //change des angles en degres, en radian
	  public static double degToRad(double valDeg){ //petite methode pour passer en radians
	  	return  valDeg / 180 * 3.141592654;
	  }
	  
	  //calcul la distance entre deux sommets
	  public static double distance(Sommet un,Sommet deux){
		  return degreesToKm(un.getLatitude(),un.getLongitude(),deux.getLatitude(),deux.getLongitude());
	  }
  
	
	
	public static int isConnexe(int[] tabConnexite){
		int nbComp = 0;
		for(int i=0; i < tabConnexite.length; i++)	{
			if (tabConnexite[i] > nbComp )
				nbComp = tabConnexite[i];
		}
		return nbComp;
	}
	
	public static void printTab(int[] tab){
		String result = "";
			
		for(int i=0 ; i < tab.length; i++){
			result =  "Sommet " + (i) + " : ";
			result = result + tab[i] + "\n";
			System.out.println(result);
		}
	}
	
	/**
	 * Les méthodes calculConnexite et testConnexite sont a ameliorer ! pour eliminer la recursivite (qui amene un stackoverflow sur la carte d'europe)
	 * 
	 */
	public static int[] calculConnexite(Graphe graph){
		int nbSommets = graph.getNbSommet();
		int[] tableauConnexite = new int[nbSommets];
		int i=0;
		int numGpeConnexe = 1;
		while(i<nbSommets){
			if((tableauConnexite[i] == 0) && (graph.getListeSommets().get(i) != null)){
				testConnexite(tableauConnexite, ((Sommet) graph.getListeSommets().get(i)).getIdSom(), numGpeConnexe, graph);
				numGpeConnexe++;
			}

			i++;
		}

		return tableauConnexite;
	}
	
	private static void testConnexite (int[] tableauConnexite, int sommet, int numGpeConnexe, Graphe graph){
		
		//System.out.println("set sommet "+sommet+" a "+numGpeConnexe);
		
		if(tableauConnexite[sommet] == 0) {
			tableauConnexite[sommet] = numGpeConnexe;
			ArrayList<Successeur> successeurs = graph.getListeSucc().get(sommet);
			if (!successeurs.isEmpty()){
	
				int j=0;
			//	System.out.println("j = "+j);
			//	System.out.println("nb successeurs = " + successeurs.size());
			//	System.out.println(successeurs);
				while ((j < successeurs.size()) && (successeurs.get(j) != null)) {
					
				//	System.out.println("demande set sommet "+((Successeur) successeurs.get(j)).getNumSommet());
	
					testConnexite(tableauConnexite, ((Successeur) successeurs.get(j)).getNumSommet(), numGpeConnexe, graph);
					j++;
					
				}
			}
		}
	}
	
	public static void construirePetitGraphe(Graphe graph,petitGraphe petitgraph,int[] tabComposantes){
		ArrayList<Sommet> listeStations = petitgraph.getListeStations();
		Trajet trajet;
		int numDeb;
		int numFin;
		
		for(int i=0; i < listeStations.size(); i++){ //pour toutes les stations
			numDeb = listeStations.get(i).getIdSom();
			for(int j=0;j < listeStations.size();j++){
				numFin = listeStations.get(j).getIdSom();
				
				//	s'il n'existe pas un trajet entre les deux stations qu'elles sont connexes, et que la distance entre elles (a vol d'oiseau) soit inférieure à 150 Km, on calcul un plus court chemin
				if((numFin != numDeb) && !petitgraph.trajetExiste(numDeb,numFin) && verifConnexite(numDeb, numFin, tabComposantes) && distance(listeStations.get(i),listeStations.get(j))<=150){ 
					trajet =  Algo.petitDijsktra(numDeb,numFin,graph,"distance");
					if(trajet.getDistance()<=150) { //on enregistre les trajets inferieurs a 150 Km
						petitgraph.ajouterTrajet(trajet,numDeb);
						petitgraph.ajouterTrajet(trajet,numFin);
					}
				}		
			}
		}
	}
	

	
	
	
	/****PARTIE PLUS COURTS CHEMINS *********/
	private static boolean verifConnexite(int somDepart, int somFin, int[] tabComposantes){
		return (tabComposantes[somDepart] == tabComposantes[somFin]);
	}
	
	
	public static Trajet grandDijsktra(int somDepart, int somFin, Graphe graph,String param){
			
			Trajet trajet = new Trajet();
			int sommet_min;
			double valeur_min;
			double valeur;
			int sommetTravail;
			
			int[] chemin = new int[graph.getNbSommet()];
			
			//tableau contenant pour chaque sommet du graphe, un boolean et la valeur calculee
			tabDij tab =  new tabDij(graph.getNbSommet());
			
			sommetTravail = somDepart; //pour la premiere iteration
			while(!tab.getMarqueAt(somFin)){ //tant que le sommet qui nous interesse n'est pas marque, on continue
				//on prend les successeurs de notre sommet
				ArrayList<Successeur> succ = graph.getListeSucc().get(sommetTravail);
		
				//on met a jour les distances des successeurs non marques a true
				if(!succ.isEmpty()){ //si y a des successeurs
				
					for(int j = 0; j < succ.size(); j++){
						if(!tab.getMarqueAt(((Successeur) succ.get(j)).getNumSommet())){ //si le successeur n'a pas ete marque
							Successeur leSucc = (Successeur) succ.get(j);
							int numSommet = leSucc.getNumSommet();
							
							if(param.equals("distance"))
								valeur = leSucc.getValuationArc().getDistance() + tab.getValeurAt(sommetTravail);
							else if(param.equals("temps"))
								valeur = leSucc.getValuationArc().getDistance(); //il faut calculer le temps de parcours !!
							else valeur = -1;
							
							if(((tab.getValeurAt(numSommet) == 0) || (tab.getValeurAt(numSommet) > valeur)) && (numSommet != sommetTravail)) {//si le sommet est marque avec une valeur superieure, on l'a met a jour
								tab.setValeurAt(numSommet,valeur);
								chemin[numSommet]=sommetTravail; //on sauvegarde le sommet de travail comme pr?decesseur direct du sommet courant
							}	
						}
					}			
				}
	            //on marque le sommet lui mm
					tab.setMarqueAt(sommetTravail);

				//selection du prochain sommet a marquer -- on le selectionne parmis les sommets non marques mais ayant deja une distance allou?e
					valeur_min = 0;
					sommet_min = 0;
					for(int i = 0;i<tab.getLenght();i++){
						if((!tab.getMarqueAt(i) && tab.getValeurAt(i) > 0) && (valeur_min == 0 || tab.getValeurAt(i) < valeur_min)){
							valeur_min = tab.getValeurAt(i);
							sommet_min = i;
						}
					}
				//on passe au prochain sommet
					sommetTravail = sommet_min;
					

				
			}//Fin du marquage du graphe
			
					
				sommetTravail = somFin;
				trajet.addEtape(sommetTravail);
				while(sommetTravail != somDepart ){
					trajet.addEtape(chemin[sommetTravail]);
					sommetTravail = chemin[sommetTravail];
				}
				trajet.setDistance(tab.getValeurAt(somFin));
				return trajet;
				
	}
	
	public static Trajet petitDijsktra(int somDepart, int somFin, Graphe graph,String param){
		
			Trajet trajet = new Trajet();
			int[] chemin = new int[graph.getNbSommet()+1];
			int auxNum;
			int sommetTravail = somDepart;
			InfoTas infoTravail = new InfoTas(somDepart,0);
			InfoTas infoTemp;
			double coutTemp;
			
			Vtas tas = new Vtas(graph.getNbSommet());
			
			tas.inserer(somDepart,0); //on insere le premier sommet dans le tas, pour eviter des pb de pointeur null apres
			while(sommetTravail != somFin){
				 
				ArrayList<Successeur>  succ = graph.getListeSucc().get(sommetTravail); //on recupere les successeurs du sommet de travail
				
				for(int i=0;i<succ.size();i++){ //pour chaque successeur
					auxNum = succ.get(i).getNumSommet();
				//	System.out.println("Iteration, on est au sommet "+ sommetTravail + " (soit "+ graph.nom(sommetTravail)+")");
					if(auxNum != sommetTravail && !tas.marque(auxNum)) {//on evite les boucles (y en a dans les fichiers
						if(!tas.appartient(auxNum)) //on regarde s'il est deja dans le tas
							tas.inserer(auxNum,(float) Integer.MAX_VALUE); //on l'insere si besoin
							
						infoTemp = tas.getSommet(auxNum); //on recupere les infos du successeur
						coutTemp = infoTravail.getCout()+succ.get(i).getValuationArc().getDistance(); // on calcul le cout temporaire si on prend le chemin venant du sommet de travail
						
						if ((infoTemp.getCout() > coutTemp) && chemin[sommetTravail] != auxNum) { //s'il le faut, on met a jour
							tas.maj(auxNum,(float) coutTemp);
							chemin[auxNum]=sommetTravail;
						}
					}
				}
				tas.retirerTete(); //on retire le sommet traite du tas
				infoTravail = tas.getTete(); //on recupere le sommet suivant
				sommetTravail = infoTravail.getIdSom(); //on met de cote son numero, pour y acceder plus vite 						
			}
			
//			le marquage est termine, on peut recuperer le plus court chemin
				
			sommetTravail = somFin;
			trajet.addEtape(sommetTravail);
			while(sommetTravail != somDepart ){
				trajet.addEtape(chemin[sommetTravail]);
				sommetTravail = chemin[sommetTravail];
			}
			trajet.setDistance(tas.getSommet(somFin).getCout());
			return trajet;					
	}
	
	
	public static Trajet dijsktraContraint(int somDepart, int somFin,Graphe gdGraph, petitGraphe graph,String param){
		
		//operation preliminaire : inserer les deux sommets (si besoin) dans le petit graph et leur ajouter des arcs
		
		Sommet somTemp;
		ArrayList<Sommet> ls = graph.getListeStations();
		
		if(graph.getId(somDepart)==-1){
			graph.insererSommet(somDepart);
			somTemp = gdGraph.getListeSommets().get(somDepart);
			for(int i=0;i<graph.getNbSommet();i++){
				if(distance(somTemp,ls.get(i)) < 150){ //on considere que si une ville est situee a moins de 150Km d'une station, alors il y a un trajet entre ces deux villes 
					Trajet temp = new Trajet();
					temp.addEtape(somTemp.getIdSom());
					temp.addEtape(ls.get(i).getIdSom());
					temp.setDistance(distance(somTemp,ls.get(i)));
					graph.ajouterTrajet(temp,somTemp.getIdSom());
					
				}
			}
		}
		
		if(graph.getId(somFin)==-1){
			graph.insererSommet(somFin);
			somTemp = gdGraph.getListeSommets().get(somFin);
			for(int i=0;i<graph.getNbSommet();i++){
				if(distance(somTemp,ls.get(i)) < 150){
					Trajet temp = new Trajet();
					temp.addEtape(somTemp.getIdSom());
					temp.addEtape(ls.get(i).getIdSom());
					temp.setDistance(distance(somTemp,ls.get(i)));
					graph.ajouterTrajet(temp,somTemp.getIdSom());
				}
			}
		}

		
		Trajet trajet = new Trajet();
		int[] chemin = new int[graph.getNbSommet()+1];
		int auxNum;
		int sommetTravail = graph.getId(somDepart);
		int numDeb = sommetTravail;
		int numFin = graph.getId(somFin);
		InfoTas infoTravail = new InfoTas(sommetTravail,0);
		InfoTas infoTemp;
		double coutTemp;
		
		Vtas tas = new Vtas(graph.getNbSommet());
		
		tas.inserer(sommetTravail,0); //on insere le premier sommet dans le tas, pour eviter des pb de pointeur null apres
		while(sommetTravail != numFin){
			 
			ArrayList<Trajet>  traj = graph.getListeTrajets().get(sommetTravail); //on recupere les trajets qui partent du sommet de travail
			for(int i=0;i<traj.size();i++){ //pour chaque successeur
				auxNum = (sommetTravail==graph.getId(traj.get(i).getDeb())) ? graph.getId(traj.get(i).getFin()) : graph.getId(traj.get(i).getDeb());
			//	System.out.println("Iteration, on est au sommet "+ sommetTravail + " (soit "+ graph.nom(sommetTravail)+")");
				if(auxNum != sommetTravail && !tas.marque(auxNum)) {//on evite les boucles (y en a dans les fichiers
					if(!tas.appartient(auxNum)) //on regarde s'il est deja dans le tas
						tas.inserer(auxNum,(float) Integer.MAX_VALUE); //on l'insere si besoin
						
					infoTemp = tas.getSommet(auxNum); //on recupere les infos du successeur
					coutTemp = infoTravail.getCout()+traj.get(i).getDistance(); // on calcul le cout temporaire si on prend le chemin venant du sommet de travail
					
					if ((infoTemp.getCout() > coutTemp) && chemin[sommetTravail] != auxNum) { //s'il le faut, on met a jour
						tas.maj(auxNum,(float) coutTemp);
						chemin[auxNum]=sommetTravail;
					}
				}
			}
			if(tas.taille()==0)
				{
				return null;
				}
			tas.retirerTete(); //on retire le sommet traite du tas
			infoTravail = tas.getTete(); //on recupere le sommet suivant
			sommetTravail = infoTravail.getIdSom(); //on met de cote son numero, pour y acceder plus vite 						
		}
		
//		le marquage est termine, on peut recuperer le plus court chemin
			
		sommetTravail = numFin;
		trajet.addEtape(graph.getNum(sommetTravail));
		while(sommetTravail != numDeb ){
			trajet.addEtape(graph.getNum(chemin[sommetTravail]));
			sommetTravail = chemin[sommetTravail];
		}
		trajet.setDistance(tas.getSommet(numFin).getCout());
		return trajet;

		}
	
	
	
	//affiche un trajet
	public static String afficheTrajet(Trajet t, Graphe g){
		String retour = "";
		while(!t.isEmpty())
			retour += g.nom(t.etape())+"\n";
		retour += t.taille()+" Etapes - "+t.getDistance()+"Km\n";
		return retour;
	}
	
	
	//s'occupe de gerer les pb et de calculer le temps (et de lancer les bons algo de plus court chemin)
	public static void plusCourtChemin(String depart, String fin, Graphe graph,petitGraphe petitgraph,String param, int[] tabComposantes, String choix){
		int numDeb = graph.num(depart);
		int numFin = graph.num(fin);
		Trajet trajet;
		String affichage="";
		if(numDeb == -1) 
			System.out.println(depart+" n'est pas dans le graphe");
		else if (numFin == -1)
			System.out.println(fin+" n'est pas dans le graphe");
		else if(!verifConnexite(numDeb,numFin,tabComposantes))
			System.out.println("Les deux sommets ne sont pas dans la meme composante connexe");
		else {
			Date tDeb = new Date();
			if(choix.equals("tas"))
				 trajet =  Algo.petitDijsktra(numDeb,numFin,graph,param); //Dijsktra implémentant le tas
			else if(choix.equals("normal"))
				 trajet =  Algo.grandDijsktra(numDeb,numFin,graph,param); //Dijsktra simple
			else 
				 trajet = Algo.dijsktraContraint(numDeb,numFin,graph,petitgraph,param); //Dijstra + tas + contrainte de 150Km 
			Date tFin = new Date();
			affichage = (trajet!=null) ? Algo.afficheTrajet(trajet,graph): "Il n'y a pas de solution !";				
			System.out.println(affichage);
			System.out.println("Le calcul a dure " + (tFin.getTime()-tDeb.getTime()) + "ms"); 
		}
			
		
	}
}

